from sqlalchemy import JSON, BigInteger, Column, DateTime, String
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class InferenceLog(Base):
    __tablename__ = "mlfoundry_inference_log"
    id = Column(BigInteger, primary_key=True, autoincrement=True)

    # Identifying the model
    model_name = Column(String(256), nullable=False)
    model_version = Column(String(64), nullable=False)
    # identifying the inference payload
    inference_id = Column(String(36), nullable=False)
    features = Column(JSON, nullable=False)
    predictions = Column(JSON, nullable=False)
    occurred_at = Column(DateTime(), nullable=False)

    shap_values = Column(JSON)
    raw_data = Column(JSON)
    actuals = Column(JSON)



import enum


class FileFormat(enum.Enum):
    CSV = "csv"
    PARQUET = "parquet"


class ModelFramework(enum.Enum):
    SKLEARN = "sklearn"
    TENSORFLOW = "tensorflow"
    PYTORCH = "pytorch"
    KERAS = "keras"
    XGBOOST = "xgboost"
    LIGHTGBM = "lightgbm"
    FASTAI = "fastai"
    H2O = "h2o"
    ONNX = "onnx"
    SPACY = "spacy"
    STATSMODELS = "statsmodels"
    GLUON = "gluon"
    PADDLE = "paddle"


class DataSlice(enum.Enum):
    TRAIN = "train"
    VALIDATE = "validate"
    TEST = "test"
    PREDICTION = "prediction"


class ModelType(enum.Enum):
    BINARY_CLASSIFICATION = "binary_classification"
    MULTICLASS_CLASSIFICATION = "multiclass_classification"
    REGRESSION = "regression"
    TIMESERIES = "timeseries"



import typing

import orjson
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import scoped_session, sessionmaker

from ..inference_packet import InferencePacket
from .abstract_inference_store import AbstractInferenceStore
from .db_models import Base, InferenceLog


def orjson_serializer(obj):
    return orjson.dumps(obj, option=orjson.OPT_SERIALIZE_NUMPY).decode()


def orjson_deserializer(serialized_obj):
    return orjson.loads(serialized_obj)


def create_sqlalchemy_engine(db_uri: str) -> Engine:
    engine = create_engine(
        db_uri, json_serializer=orjson_serializer, json_deserializer=orjson_deserializer
    )
    return engine


# NOTE: for now I am keeping the above helper function in this file,
# later we may need to create a seperate db util file to keep this functions


class SqlAlchemyInferenceStore(AbstractInferenceStore):
    def __init__(self, db_uri: str):
        self.engine: Engine = create_sqlalchemy_engine(db_uri)

        # TODO: Explore albemic later
        Base.metadata.create_all(self.engine)
        self.session = scoped_session(sessionmaker(self.engine))

    def batch_log_inference(self, inference_packets: typing.List[InferencePacket]):
        with self.session() as session, session.begin():
            # https://github.com/sqlalchemy/sqlalchemy/issues/6519
            inference_log_objects = [
                InferenceLog(
                    model_name=inference.model_name,
                    model_version=inference.model_version,
                    inference_id=inference.inference_id,
                    features=inference.features,
                    predictions=inference.predictions,
                    occurred_at=inference.occurred_at,
                    raw_data=inference.raw_data,
                    actuals=inference.actuals,
                    shap_values=inference.shap_values,
                )
                for inference in inference_packets
            ]
            session.add_all(inference_log_objects)



from mlfoundry.frameworks.fastai_registry import FastAIRegistry
from mlfoundry.frameworks.gluon_registry import GluonRegistry
from mlfoundry.frameworks.h2o_registry import H2ORegistry
from mlfoundry.frameworks.keras_registry import KerasRegistry
from mlfoundry.frameworks.lightgbm_registry import LightGBMRegistry
from mlfoundry.frameworks.onnx_registry import ONNXRegistry
from mlfoundry.frameworks.paddle_registry import PaddleRegistry
from mlfoundry.frameworks.pytorch_registry import PyTorchRegistry
from mlfoundry.frameworks.sklearn_registry import SkLearnRegistry
from mlfoundry.frameworks.spacy_registry import SpacyRegistry
from mlfoundry.frameworks.statsmodel_registry import StatsModelsRegistry
from mlfoundry.frameworks.tensorflow_registry import TensorflowRegistry
from mlfoundry.frameworks.xgboost_registry import XGBoostRegistry



import mlflow

from mlfoundry.exceptions import MlflowException, MlFoundryException


def download_artifact(mlflow_client, run_id, artifact_name, dest_path):
    try:
        return mlflow_client.download_artifacts(run_id, artifact_name, dest_path)
    except MlflowException as e:
        raise MlFoundryException(e.message).with_traceback(e.__traceback__) from None



from concurrent.futures import ThreadPoolExecutor
from threading import BoundedSemaphore


class AsyncExecuter:
    """
    BoundedExecutor behaves as a ThreadPoolExecutor which will block on
    calls to submit() once the limit given as "bound" work items are queued for
    execution.

    :param bound: Integer - the maximum number of items in the work queue
    :param max_workers: Integer - the size of the thread pool
    """

    def __init__(self, bound=100, max_workers=8):
        self.executor = ThreadPoolExecutor(max_workers=max_workers)
        self.semaphore = BoundedSemaphore(bound + max_workers)

    def submit(self, fn, *args, **kwargs):
        self.semaphore.acquire()
        try:
            future = self.executor.submit(fn, *args, **kwargs)
        except:
            self.semaphore.release()
            raise
        else:
            future.add_done_callback(lambda x: self.semaphore.release())
            return future

    def shutdown(self, wait=False):
        self.executor.shutdown(wait)



import abc
import typing

from ..inference_packet import InferencePacket


class AbstractInferenceStore(abc.ABC):
    def log_inference(self, inference_packet: InferencePacket):
        self.batch_log_inference([inference_packet])

    @abc.abstractmethod
    def batch_log_inference(self, inference_packets: typing.List[InferencePacket]):
        ...
